// $Header: /home/cvs/jakarta-jmeter/src/core/org/apache/jmeter/visualizers/SamplingStatCalculator.java,v 1.7 2005/07/12 20:51:00 mstover1 Exp $
/*
 * Copyright 2001-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */

package se.six.jmeter.visualizer.statagg;

import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.samplers.StatisticalSampleResult;
import org.apache.jmeter.visualizers.Sample;
import org.apache.jorphan.logging.LoggingManager;
import org.apache.log.Logger;

import java.io.Serializable;
import java.text.DecimalFormat;

/**
 * Aggegate sample data container. Just instantiate a new instance of this
 * class, and then call {@link #addSample(org.apache.jmeter.samplers.SampleResult)} a few times, and pull
 * the stats out with whatever methods you prefer.
 *
 * @author James Boutcher
 * @version $Revision: 1.7 $
 */
public class StatAggSampler
  implements Serializable
{
  static Logger log = LoggingManager.getLoggerForClass();

  private static DecimalFormat rateFormatter = new DecimalFormat("#.0");

  private static DecimalFormat errorFormatter = new DecimalFormat("#0.00%");

  private static DecimalFormat kbFormatter = new DecimalFormat("#0.00");

  private StatAggCalculator calculator = new StatAggCalculator();

  //private ArrayList storedValues = new ArrayList();
  private Sample currentSample;

  private double maxThroughput;

  private long firstTime;

  private String label;

  // private int index;

  private StatAggSampler()
  {// Don't (can't) use this...
  }

  /**
   * Use this constructor.
   */
  public StatAggSampler(String label)
  {
    this.label = label;
    init();
  }

  /**
   * Essentially a copy function
   *
   * @param stat
   */
  /*public StatAggSampler(StatAggSampler stat) {
    this(stat.label);
    addSamples(stat);
  }*/

  private void init()
  {
    firstTime = Long.MAX_VALUE;
    calculator.clear();
    //storedValues.clear();
    maxThroughput = Double.MIN_VALUE;
  }

  /*public void addSamples(StatAggSampler ssc) {
    calculator.addAll(ssc.calculator);
    //storedValues.addAll(ssc.storedValues);
    //Collections.sort(storedValues);
    if (firstTime > ssc.firstTime) {
      firstTime = ssc.firstTime;
    }
  }*/

  /**
   * Clear the counters (useful for differential stats)
   */
  public synchronized void clear()
  {
    init();
  }

  public Sample getCurrentSample()
  {
    /*if (storedValues.size() == 0) {
      return new Sample();
    }
    return (Sample) storedValues.get(storedValues.size() - 1);*/

    if (currentSample == null) {
      return new Sample();
    }

    return currentSample;
  }

  /**
   * Get the elapsed time for the samples
   *
   * @return how long the samples took
   */
  public long getElapsed()
  {
    if (getCurrentSample().endTime == 0)
      return 0;// No samples collected ...
    return getCurrentSample().endTime - firstTime;
  }

  /**
   * Returns the throughput associated to this sampler in requests per second.
   * May be slightly skewed because it takes the timestamps of the first and
   * last samples as the total time passed, and the test may actually have
   * started before that start time and ended after that end time.
   */
  public double getRate()
  {
    if (calculator.getCount() == 0)
      return 0.0; // Better behaviour when howLong=0 or lastTime=0

    return getCurrentSample().throughput;
  }

  /**
   * Returns a String that represents the throughput associated for this
   * sampler, in units appropriate to its dimension:
   * <p/>
   * The number is represented in requests/second or requests/minute or
   * requests/hour.
   * <p/>
   * Examples: "34.2/sec" "0.1/sec" "43.0/hour" "15.9/min"
   *
   * @return a String representation of the rate the samples are being taken
   *         at.
   */
  public String getRateString()
  {
    double rate = getRate();

    if (rate == Double.MAX_VALUE) {
      return "N/A";
    }

    String unit = "sec";

    if (rate < 1.0) {
      rate *= 60.0;
      unit = "min";
    }
    if (rate < 1.0) {
      rate *= 60.0;
      unit = "hour";
    }

    String rval = rateFormatter.format(rate) + "/" + unit;

    return (rval);
  }

  /**
   * calculates the average page size, which means divide the bytes by number
   * of samples.
   *
   * @return
   */
  public double getPageSize()
  {
    double rate = 0;
    if (this.getElapsed() > 0 && calculator.getTotalBytes() > 0) {
      rate = calculator.getTotalBytes() / ((double)this.getElapsed() / 1000);
    }
    if (rate < 0) {
      rate = 0;
    }
    return rate;
  }

  /**
   * formats the rate
   *
   * @return
   */
  public String getPageSizeString()
  {
    double rate = getPageSize() / 1024;
    return kbFormatter.format(rate);
  }

  public String getLabel()
  {
    return label;
  }

  /**
   * Records a sample.
   */
  public Sample addSample(SampleResult res)
  {

    synchronized (calculator) {
      long byteslength = 0;
      // in case the sampler doesn't return the contents
      // we see if the bytes was set
      if (res.getResponseData() == null || res.getResponseData().length == 0) {
        byteslength = res.getBytes();
      } else {
        byteslength = res.getResponseData().length;
      }
      // if there was more than 1 loop in the sample, we
      // handle it appropriately
      if (res.getSampleCount() > 1) {
        long time = res.getTime() / res.getSampleCount();
        long resbytes = byteslength / res.getSampleCount();
        for (int idx = 0; idx < res.getSampleCount(); idx++) {
          calculator.addValue(time);
          calculator.addBytes(resbytes);
        }
      } else {
        calculator.addValue(res.getTime());
        calculator.addBytes(byteslength);
      }
      setStartTime(res);
      long eCount = getCurrentSample().errorCount;
      if (res instanceof StatisticalSampleResult) {
        eCount += ((StatisticalSampleResult)res).getErrorCount();
      } else if (!res.isSuccessful()) {
        eCount++;
      }
      long endTime = getEndTime(res);
      long howLongRunning = endTime - firstTime;
      double throughput = 0;
      if (howLongRunning <= 0) {
        throughput = Double.MAX_VALUE;
      }
      throughput = ((double)calculator.getCount() / (double)howLongRunning) * 1000.0;
      if (throughput > maxThroughput) {
        maxThroughput = throughput;
      }
      currentSample = new Sample(null,
                                 res.getTime(),
                                 (long)calculator.getMean(),
                                 (long)calculator.getStandardDeviation(),
                                 calculator.getMedian().longValue(),
                                 calculator.getPercentPoint(0.500).longValue(),
                                 throughput,
                                 eCount,
                                 res.isSuccessful(),
                                 calculator.getCount(),
                                 endTime);
      //storedValues.add(s);
      //System.out.println("SampleEvent: '" + res.getSampleLabel() + "', Bytes: " + byteslength + ",Count: " + res.getSampleCount() + ", Errors: " + eCount);
    }

    return currentSample;
  }

  /*public List getSamples() {
    return storedValues;
  }

  public Sample getSample(int index) {
    if (index < storedValues.size()) {
      return (Sample) storedValues.get(index);
    }
    return null;
  }*/

  private long getEndTime(SampleResult res)
  {
    long endTime = res.getTimeStamp();
    long lastTime = getCurrentSample().endTime;
    if (res.isStampedAtStart()) {
      endTime += res.getTime();
    }
    if (lastTime < endTime) {
      lastTime = endTime;
    }
    return lastTime;
  }

  /**
   * @param res
   */
  private void setStartTime(SampleResult res)
  {
    long startTime = res.getTimeStamp();
    if (!res.isStampedAtStart()) {
      startTime -= res.getTime();
    }
    if (firstTime > startTime) {
      // this is our first sample, set the start time to current timestamp
      firstTime = startTime;
    }
  }

  /**
   * Returns the raw double value of the percentage of samples with errors
   * that were recorded. (Between 0.0 and 1.0) If you want a nicer return
   * format, see {@link #getErrorPercentageString()}.
   *
   * @return the raw double value of the percentage of samples with errors
   *         that were recorded.
   */
  public double getErrorPercentage()
  {
    double rval = 0.0;

    if (calculator.getCount() == 0) {
      return (rval);
    }
    rval = (double)getCurrentSample().errorCount / (double)calculator.getCount();
    return (rval);
  }

  /**
   * Returns a String which represents the percentage of sample errors that
   * have occurred. ("0.00%" through "100.00%")
   *
   * @return a String which represents the percentage of sample errors that
   *         have occurred.
   */
  public String getErrorPercentageString()
  {
    double myErrorPercentage = this.getErrorPercentage();
    if (myErrorPercentage < 0) {
      myErrorPercentage = 0.0;
    }

    return (errorFormatter.format(myErrorPercentage));
  }

  /**
   * For debugging purposes, mainly.
   */
  public String toString()
  {
    StringBuffer mySB = new StringBuffer();

    mySB.append("Samples: " + this.getCount() + "  ");
    mySB.append("Avg: " + this.getMean() + "  ");
    mySB.append("Min: " + this.getMin() + "  ");
    mySB.append("Max: " + this.getMax() + "  ");
    mySB.append("Error Rate: " + this.getErrorPercentageString() + "  ");
    mySB.append("Sample Rate: " + this.getRateString());
    return (mySB.toString());
  }

  /**
   * @return errorCount
   */
  public long getErrorCount()
  {
    return getCurrentSample().errorCount;
  }

  /**
   * @return Returns the maxThroughput.
   */
  public double getMaxThroughput()
  {
    return maxThroughput;
  }


  /**
   * @param percent
   * @return
   */
  public Number getPercentPoint(double percent)
  {
    return calculator.getPercentPoint(percent);
  }

  /**
   * @return
   */
  public int getCount()
  {
    return calculator.getCount();
  }

  /**
   * @return
   */
  public Number getMax()
  {
    return calculator.getMax();
  }

  /**
   * @return
   */
  public double getMean()
  {
    return calculator.getMean();
  }

  public Number getMeanAsNumber()
  {
    return new Long((long)calculator.getMean());
  }

  /**
   * @return
   */
  public Number getMedian()
  {
    return calculator.getMedian();
  }

  /**
   * @return
   */
  public Number getMin()
  {
    if (calculator.getMin().longValue() < 0) {
      return new Long(0);
    } else {
      return calculator.getMin();
    }
  }

  /**
   * @param percent
   * @return
   */
  public Number getPercentPoint(float percent)
  {
    return calculator.getPercentPoint(percent);
  }

  /**
   * @return
   */
  public double getStandardDeviation()
  {
    return calculator.getStandardDeviation();
  }

  public Number getStandardDeviationAsNumber()
  {
    return Long.valueOf((long)calculator.getStandardDeviation());
  }
} // class RunningSample
